/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#ifndef _DOC_H_
#define _DOC_H_

#include "cursor.h"
#include "options.h"
#include "play.h"
#include "history.h"
#include "midi.h"

#include <stdio.h>

typedef enum {
  NOTE, SILENCE
} elem_type;

typedef enum {
  NORMAL, DIESE, BEMOL, BECARRE
} note_alteration;

typedef enum {
  RONDE, BLANCHE, NOIRE, CROCHE, DCROCHE, DDCROCHE, DDDCROCHE, DDDDCROCHE
} note_duration;

typedef struct {
#if 0
  int width;
  int height;
  int center_x;
  int center_y;
#endif
  int max_x;
  int min_x;
  int max_y;
  int min_y;
  int stem_min_y;
  int stem_max_y;
  note_duration stem_type;
} elem_dimension;

typedef struct {
  elem_type type;
  union {
    struct {
      int pitch;	/* 0 = middle G */
      note_alteration alteration;
      note_duration duration;
      int nb_dots;
      int head_delta;
      int alteration_delta;
      int dots_delta;
      int stem_down:1;
    } note;
    struct {
      int position;	/* 0 = middle G */
      note_duration duration;
      int nb_dots;
    } silence;
  } u;
  /* drawing stuff */
  elem_dimension dim;	/* relative to the center of the note/silence */
} elem;

typedef enum {
  NOTES_CHORD, MESURE_BREAK, TONALITE, CLEF
} chord_type;

typedef enum {
  CLEF_G, CLEF_F, CLEF_C, CLEF_G_8va, CLEF_G_8vb
} clef_type;

typedef struct {
  int max_x;
  int min_x;
  int max_y;
  int min_y;
  int width;
  int x;
  int stem_up_min_y;
  int stem_up_max_y;
  note_duration stem_up_type; /* test if RONDE to not put a stem */
  int stem_down_min_y;
  int stem_down_max_y;
  note_duration stem_down_type; /* test if RONDE to not put a stem */
} chord_dimension;

typedef struct {
  chord_type type;
  union {
    struct {
      elem *data;
      int size;
      int maxsize;
    } chord;
    struct {
      int position;	/* 0 = middle G */
      note_alteration alteration;  /* NORMAL not allowed */
    } tonalite;
    struct {
      clef_type type;
      int position;	/* 0 = middle G */
    } clef;
  } u;
  /* drawing stuff */
  chord_dimension dim;	/* relative to the upper first line of the staff for y
			 * and the center of notes for x
                         */
} chord;

typedef enum {
  LINE_MUSIC, LINE_TEXT
} line_type;

typedef struct {
  int width;
  int y;	/* relative to the page */
  int height;
  /* relative to the first upper line */
  int min_y;
  int max_y;
} line_dimension;

typedef struct {
  line_type type;
  union {
    struct {
      chord *data;
      int size;
      int maxsize;
    } music;
    struct {
      char *data;
    } text;
  } u;
  /* drawing stuff */
  line_dimension dim;
  int next_invalid; /* -2: not invalid, -1: end of list, >=0: the next */
} line;

#include "select.h"

typedef enum {
  STATE_NEW, STATE_SAVED, STATE_MODIFIED
} document_state;

typedef struct doc {
  line *data;
  int size;
  int maxsize;
  /* gui stuff */
  void *drawing_area;
  void *tempo_text;
  void *fonts;
  void *desc_label;
  cursor cursor;
  int width;
  int height;
  int invalid_line; /* -1 if nothing invalid, >=0 first line invalid */
  int deleted_line; /* -1 if no deleted line, >=0 first deleted line */
  int last_positioned_line;
  /* those two to manage scrollbars */
  int current_width;
  int current_height;
  /* this new ones to not resize too often (it seems to be costly) */
  int allocated_width;
  int allocated_height;
  /* play stuff */
  play p;
  int tempo;
  int line_in_play;
  int chord_in_play;
  midi midi;
  /* options */
  options options;
  /* file name */
  char filename[FILENAME_MAX];
  /* selection stuff */
  selection sel;
  copy_buffer sel_buf;
  /* state of the doc */
  document_state state;
  /* history stuff */
  history history;
  /* zoom stuff */
  int zoom;
} doc;

#define LINE(d, i) (d)->data[i]
#define LINE_TYPE(d, i) (d)->data[i].type
#define LINE_MUSIC_DATA(d, i) (d)->data[i].u.music.data
#define LINE_MUSIC_SIZE(d, i) (d)->data[i].u.music.size
#define LINE_MUSIC_MAXSIZE(d, i) (d)->data[i].u.music.maxsize
#define LINE_TEXT_DATA(d, i) (d)->data[i].u.text.data

#define CHORD(d, l, i) LINE(d, l).u.music.data[i]
#define CHORD_TYPE(d, l, i) CHORD(d, l, i).type
#define CHORD_DATA(d, l, i) CHORD(d, l, i).u.chord.data
#define CHORD_SIZE(d, l, i) CHORD(d, l, i).u.chord.size
#define CHORD_MAXSIZE(d, l, i) CHORD(d, l, i).u.chord.maxsize
#define TONALITE_POSITION(d, l, i) CHORD(d, l, i).u.tonalite.position
#define TONALITE_ALTERATION(d, l, i) CHORD(d, l, i).u.tonalite.alteration
#define CLEF_TYPE(d, l, i) CHORD(d, l, i).u.clef.type
#define CLEF_POSITION(d, l, i) CHORD(d, l, i).u.clef.position
#define CHORD_DIM(d, l, i) CHORD(d, l, i).dim

#define NOTE(d, l, c, i) CHORD(d, l, c).u.chord.data[i]
#define NOTE_TYPE(d, l, c, i) NOTE(d, l, c, i).type
#define NOTE_PITCH(d, l, c, i) NOTE(d, l, c, i).u.note.pitch
#define NOTE_ALTERATION(d, l, c, i) NOTE(d, l, c, i).u.note.alteration
#define NOTE_DURATION(d, l, c, i) NOTE(d, l, c, i).u.note.duration
#define NOTE_HEAD_DELTA(d, l, c, i) NOTE(d, l, c, i).u.note.head_delta
#define NOTE_ALTERATION_DELTA(d, l, c, i) \
				NOTE(d, l, c, i).u.note.alteration_delta
#define NOTE_DOTS_DELTA(d, l, c, i) NOTE(d, l, c, i).u.note.dots_delta
#define NOTE_STEM_DOWN(d, l, c, i) NOTE(d, l, c, i).u.note.stem_down
#define NOTE_NB_DOTS(d, l, c, i) NOTE(d, l, c, i).u.note.nb_dots
#define SILENCE_POSITION(d, l, c, i) NOTE(d, l, c, i).u.silence.position
#define SILENCE_DURATION(d, l, c, i) NOTE(d, l, c, i).u.silence.duration
#define SILENCE_NB_DOTS(d, l, c, i) NOTE(d, l, c, i).u.silence.nb_dots
#define NOTE_DIM(d, l, c, i) NOTE(d, l, c, i).dim

typedef enum {
  MERGE_DONT_DELETE, MERGE_DELETE
} merge_enum;

int doc_insert_note(doc *, int, int, elem_type, ...);
int doc_set_note_pitch(doc *, int, int, int, int);
int doc_set_note_position(doc *, int, int, int, int);
int doc_set_note_alteration(doc *, int, int, int, note_alteration);
int doc_set_note_duration(doc *, int, int, int, note_duration);
int doc_set_note_dot(doc *, int, int, int, int);
int doc_set_note_head_delta(doc *, int, int, int, int);
int doc_set_note_alteration_delta(doc *, int, int, int, int);
int doc_set_note_dots_delta(doc *, int, int, int, int);
int doc_set_note_stem_down(doc *, int, int, int, int);
int doc_set_tonalite_position(doc *, int, int, int);
int doc_set_tonalite_alteration(doc *, int, int, note_alteration);
int doc_set_clef_type(doc *, int, int, clef_type);
int doc_set_clef_position(doc *, int, int, int);
int doc_insert_chord(doc *, int, int, chord_type);
int doc_insert_line(doc *, line_type, int);
int doc_merge_lines(doc *, int, merge_enum);
int doc_split_line(doc *, int, int, void *, int);
int doc_delete_line(doc *, int);
int doc_delete_chord(doc *, int, int);
int doc_delete_note(doc *, int, int, int);

void destroy_doc(doc *);

void destroy_line(line *);
void destroy_chord(chord *);
int copy_chord(chord *, chord *);

int get_delta_clef(doc *, int, int);

#endif /* _DOC_H_ */
