/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "error.h"

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#define BUFLEN 65536

static char buf[BUFLEN];
static int pos;

const char *get_error(void)
{
  return buf;
}

static void add_char(char c)
{
  if (pos == BUFLEN) return;
  buf[pos] = c;
  pos++;
}

static void add_string(char *s)
{
  while (*s) {
    add_char(*s);
    s++;
  }
}

static void add_endchar(void)
{
  if (pos == BUFLEN) pos--;
  buf[pos] = 0;
}

void set_error(enum error type, ...)
{
  va_list ap;

  va_start(ap, type);

  pos = 0;

  switch (type) {
    case NO_ERROR:
      add_string("No error.\n\n");
      add_string("Since you read this, there is a bug.\n");
      add_string("Contact the authors.\n");
      break;
    case OOM:
      add_string("Out of memory.\n\n");
      add_string("Gcomposer cannot allocate memory.\n");
      add_string("Save you work and quit.\n");
      add_string("You probably run too much programs.\n");
      add_string("Quit some of them and come back in gcomposer.\n");
      break;
    case STATE_POP:
      add_string("Internal error (STATE_POP).\n\n");
      add_string("Contact the authors and explain what\n");
      add_string("you did for the problem to be resolved.\n");
      add_string("In the meantime save your work and quit.\n");
      break;
    case OSS_ERROR:
      add_string("Audio error.\n\n");
      add_string("There is a problem with audio playback.\n");
      add_string("If you never used the audio on your computer\n");
      add_string("before, be sure you have an audio device.\n");
      add_string("If you have an audio device, the program may\n");
      add_string("access it in an incorrect way.\n");
      add_string("You are trying to use 'OSS' to access it.\n");
      add_string("Try 'ALSA' in the options' menu.\n");
      add_string("Or if you are sure it is 'OSS,' try to set\n");
      add_string("another device. Or maybe another program is\n");
      add_string("also accessing the audio device, in which case\n");
      add_string("quit it and try again.\n");
      add_string("If you don't know what to do, contact the\n");
      add_string("authors for them to help you.\n");
      break;
    case FILE_ERROR: {
      file_error_type t;
      char *name;
      int err_code;
      t = va_arg(ap, file_error_type);
      name = va_arg(ap, char *);
      err_code = va_arg(ap, int);
      switch (t) {
        case FILE_OPEN_ERROR:
          add_string("Error opening the file:\n");
          break;
        case FILE_CLOSE_ERROR:
          add_string("Error closing the file:\n");
          break;
        case FILE_WRITE_ERROR:
          add_string("Error writing to the file:\n");
          break;
        case FILE_READ_ERROR:
          add_string("Error reading from the file:\n");
          break;
        case FILE_LOAD_ERROR:
          add_string("Error while loading the file:\n");
          add_string(name);
      add_string("\n\nThe file was probably not produced by Gcomposer.\n\n");
          add_string("If you are sure it was, then it might be corrupted.\n");
          add_string("There is not much to do. But contact the authors,\n");
          add_string("they might help you recover some data.\n\n");
          goto end_switch;
      }
      add_string(name);
      add_string("\n\nIf you are trying to save your work, your\n");
      add_string("harddisk may be full. Try to make some room by\n");
      add_string("deleting useless files.\n\n");
      add_string("Or you may have given a wrong file name, in which\n");
      add_string("case just try again with another file name.\n\n");
      add_string("Note that some directories may be write-protected.\n");
      add_string("If it is the case you cannot create a file in those\n");
      add_string("directories.\n\n");
      add_string("You may also have a more serious problem with your\n");
      add_string("harddisk, especially if other programs also fail\n");
      add_string("manipulating files. Run a check of your harddisk to\n");
      add_string("be sure. How to do that is highly dependant on your\n");
      add_string("operating system. Contact the appropriate support.\n");
      add_string("\nHere is the description of the error\n");
      add_string("as returned by your operating system.\n");
      add_string("It may help you correct the problem.\n\n");
      add_string(strerror(err_code));
      break;
    }
    case FATAL:
      add_string("Fatal error.\n\n");
      add_string("You should not read this.\n");
      add_string("Contact the authors and explain\n");
      add_string("them what you did to read this.\n");
      add_string("\nThe program will now exit.\n");
      break;
    case MIDI_ERROR: {
      char *name = va_arg(ap, char *);
      add_string("Error with MIDI.\n\n");
      add_string("There was a problem accessing the midi device:\n");
      add_string(name);
      add_string("\n");
      add_string("Ensure that you have a working MIDI device\n");
      add_string("(by using another program that accesses it\n");
      add_string("for example).\n");
      add_string("Ensure that the settings for MIDI are correct.\n");
      add_string("Maybe another program is already accessing the\n");
      add_string("device, in which case you may need to quit this\n");
      add_string("program and try again.\n");
      add_string("If everything seems correct, contact the\n");
      add_string("authors. The MIDI support is probably wrong.\n");
      break;
    }
  }
end_switch:

  add_endchar();

  va_end(ap);
}
