/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "select.h"

#include <stdlib.h>
#include <string.h>

#include "doc.h"
#include "mem.h"
#include "actions.h"

int copy_selection(doc *d)
{
  int l;
  int c;
  chord *e;

  if (d->sel.start_line == -1) return 1;

  destroy_selection(d);

  /* slow */
  /* but we won't bother optimizing here */
  for (l = d->sel.start_line; l <= d->sel.end_line; l++) {
    switch (LINE_TYPE(d, l)) {
      case LINE_MUSIC:
        for (c = 0; c < d->data[l].u.music.size; c++) {
          if (!in_selection(d, l, c)) continue;
          e = &CHORD(d, l, c);
          if (!alloc_select(d)) return 0;
          d->sel_buf.data[d->sel_buf.size-1].type = CD_CHORD;
          if (!copy_chord(&d->sel_buf.data[d->sel_buf.size-1].u.chord, e))
            return 0;
        }
        /* we put a CD_LINE_MARKER if next line is music too */
        if (l < d->sel.end_line && LINE_TYPE(d, l+1) == LINE_MUSIC) {
          if (!alloc_select(d)) return 0;
          d->sel_buf.data[d->sel_buf.size-1].type = CD_LINE_MARKER;
        }
        break;
      case LINE_TEXT:
        if (!alloc_select(d)) return 0;
        d->sel_buf.data[d->sel_buf.size-1].type = CD_TEXT;
        d->sel_buf.data[d->sel_buf.size-1].u.text =
            strdup(LINE_TEXT_DATA(d, l));
        if (!d->sel_buf.data[d->sel_buf.size-1].u.text) return 0;
        break;
    }
  }

  return 1;
}

static int paste_chord(doc *d, chord *c)
{
  switch (c->type) {
    case NOTES_CHORD: {
      int i;

      if (!insert_chord(d, d->cursor.line,
            /* are we at the end of the line? */
            d->cursor.chord == d->data[d->cursor.line].u.music.size ?
                -1 : d->cursor.chord, NOTES_CHORD))
        return 0;

      for (i = 0; i < c->u.chord.size; i++) {
        elem *e = c->u.chord.data + i;
        switch (e->type) {
          case NOTE:
            if (!insert_note(d, d->cursor.line, d->cursor.chord,
                             NOTE, e->u.note.pitch, e->u.note.alteration,
                             e->u.note.duration, e->u.note.nb_dots,
                             e->u.note.head_delta,
                             e->u.note.alteration_delta,
                             e->u.note.dots_delta,
                             e->u.note.stem_down))
              return 0;
            break;
          case SILENCE:
            if (!insert_note(d, d->cursor.line, d->cursor.chord,
                             SILENCE, e->u.silence.position,
                             e->u.silence.duration, e->u.silence.nb_dots))
              return 0;
            break;
        }
      }

      break;
    }
    case MESURE_BREAK:
      if (!insert_chord(d, d->cursor.line,
            /* are we at the end of the line? */
            d->cursor.chord == d->data[d->cursor.line].u.music.size ?
                -1 : d->cursor.chord, MESURE_BREAK))
        return 0;
      break;
    case TONALITE:
      if (!insert_chord(d, d->cursor.line,
            /* are we at the end of the line? */
            d->cursor.chord == d->data[d->cursor.line].u.music.size ?
               -1 : d->cursor.chord, TONALITE))
        return 0;
      if (!set_tonalite_position(d, d->cursor.line, d->cursor.chord,
                                 c->u.tonalite.position))
        return 0;
      if (!set_tonalite_alteration(d, d->cursor.line, d->cursor.chord,
                                   c->u.tonalite.alteration))
        return 0;
      break;
    case CLEF:
      if (!insert_chord(d, d->cursor.line,
            /* are we at the end of the line? */
            d->cursor.chord == d->data[d->cursor.line].u.music.size ?
               -1 : d->cursor.chord, CLEF))
        return 0;
      if (!set_clef_type(d, d->cursor.line, d->cursor.chord,
                         c->u.clef.type))
        return 0;
      if (!set_clef_position(d, d->cursor.line, d->cursor.chord,
                             c->u.clef.position))
        return 0;
      break;
  }

  return 1;
}

int paste_selection(doc *d)
{
  int i;
  copy_datum *dat;

  for (i = 0; i < d->sel_buf.size; i++) {
    dat = d->sel_buf.data + i;
    switch (dat->type) {
      case CD_LINE_MARKER:
        /* line marker exists only between 2 music line, so no problem */
        if (!split_line(d, d->cursor.line, d->cursor.chord)) return 0;
        d->cursor.chord=0;
        d->cursor.line++;
        break;
      case CD_CHORD:
        /* if current line is text:
         *   - if begin of line, insert a new line before
         *   - if end of line, insert a new line after (and inc cursor.line)
         */
        if (LINE_TYPE(d, d->cursor.line) == LINE_TEXT) {
          if (d->cursor.chord == 0) {
            /* text line, cursor at front */
            if (!insert_music_line(d, d->cursor.line-1)) return 0;
            /* line does not change */
          } else {
            /* text line, cursor at end */
            if (!insert_music_line(d, d->cursor.line)) return 0;
            d->cursor.line++;
            d->cursor.chord = 0;
          }
        }
        if (!paste_chord(d, &dat->u.chord)) return 0;
        d->cursor.chord++;
        break;
      case CD_TEXT:
        /* if current line is text:
         *   - if cursor at front, insert new line before it
         *   - if cursor at end, insert new line after it
         * if current line is music:
         *   - if cursor at front, insert new line before it
         *   - if cursor at end, insert new line after it
         *   - if none of it, split and insert after first split line
         */
        if (LINE_TYPE(d, d->cursor.line) == LINE_TEXT) {
          if (d->cursor.chord == 0) {
            /* text line, cursor at front */
            if (!insert_text_line(d, d->cursor.line-1)) return 0;
            if (!set_text(d, d->cursor.line, dat->u.text)) return 0;
            /* line does not change */
          } else {
            /* text line, cursor at end */
            if (!insert_text_line(d, d->cursor.line)) return 0;
            if (!set_text(d, d->cursor.line+1, dat->u.text)) return 0;
            d->cursor.line++;
          }
        } else {
          if (d->cursor.chord == 0) {
            /* music line, cursor at front */
            if (!insert_text_line(d, d->cursor.line-1)) return 0;
            if (!set_text(d, d->cursor.line, dat->u.text)) return 0;
            /* line does not change */
          } else if (d->cursor.chord == LINE_MUSIC_SIZE(d, d->cursor.line)) {
            /* music line, cursor at end */
            if (!insert_text_line(d, d->cursor.line)) return 0;
            if (!set_text(d, d->cursor.line+1, dat->u.text)) return 0;
            d->cursor.line++;
          } else {
            /* we split the line */
            if (!split_line(d, d->cursor.line, d->cursor.chord)) return 0;
            d->cursor.chord=0;
            d->cursor.line++;
            /* now we have music line, cursor at front */
            if (!insert_text_line(d, d->cursor.line-1)) return 0;
            if (!set_text(d, d->cursor.line, dat->u.text)) return 0;
            /* line does not change */
          }
        }
        /* we point after the text */
        d->cursor.chord = 1;
    }
  }

  return 1;
}

void destroy_selection(doc *d)
{
  int i;

  for (i = 0; i < d->sel_buf.size; i++)
    if (d->sel_buf.data[i].type == CD_CHORD)
      destroy_chord(&d->sel_buf.data[i].u.chord);
    else if (d->sel_buf.data[i].type == CD_TEXT) {
      if (d->sel_buf.data[i].u.text)
        free(d->sel_buf.data[i].u.text);
      d->sel_buf.data[i].u.text = NULL;
    }

  if (d->sel_buf.maxsize) {
    free(d->sel_buf.data);
    d->sel_buf.size = 0;
    d->sel_buf.maxsize = 0;
    d->sel_buf.data = NULL;
  }
}

int is_after(int l1, int c1, int l2, int c2)
{
  return (l1 > l2) || (l1 == l2 && c1 >= c2);
}

/* return 1 if given point is in the current selection, 0 if not */
int in_selection(doc *d, int line, int chord)
{
  if (d->sel.start_line == -1) return 0;
  return is_after(line, chord, d->sel.start_line, d->sel.start_chord) &&
         !is_after(line, chord, d->sel.end_line, d->sel.end_chord);
}
