/* gcomposer: a little tool to write music.
 * http://sed.free.fr/gcomposer
 * Released in the public domain (freedom has no license).
 */

#include "size.h"

#include "doc.h"
#include "dim.h"
#include "images.h"
#include "draw.h"
#include "gtk_wrapper.h"
#include "dprintf.h"

/* we don't want to change the scrollbars too often,
 * so following method:
 * for height:
 * increment size by block of 256
 * decrement if free size > 256*3
 * for width: same but 64 instead of 256
 */
static void setup_width_and_height(doc *d)
{
  if (d->current_width > d->allocated_width) {
    /* increase */
    d->allocated_width = (d->current_width+64+63) & (~63);
  } else if (d->current_width+64*3 < d->allocated_width) {
    /* decrease */
    d->allocated_width = (d->current_width+64+63) & (~63);
  }
  if (d->current_height > d->allocated_height) {
    /* increase */
    d->allocated_height = (d->current_height+256+255) & (~255);
  } else if (d->current_height+256*3 < d->allocated_height) {
    /* decrease */
    d->allocated_height = (d->current_height+256+255) & (~255);
  }
}

void update_horizontal_scrollbar(doc *d)
{
  /* update the scrollbar so that all visible lines may be scrolled */
  int min, max;
  int first_line, last_line;
  int w = 0;
  int i;

  gtk_wrapper_get_min_max(d->drawing_area, &min, &max, d->zoom);
  get_crossed_lines(d, min, max, &first_line, &last_line);
  dprintf("update_horizontal_scrollbar first_line=%d last_line=%d\n",
          first_line, last_line);
  if (first_line == -1) return;

  for (i=first_line; i<=last_line; i++) {
    dprintf("line %d width = %d\n", i, d->data[i].dim.width);
    if (d->data[i].dim.width > w)
      w = d->data[i].dim.width;
  }

  w += LEFT_MARGIN + RIGHT_MARGIN;

  if (w == d->current_width) return;

  d->current_width = w;

  setup_width_and_height(d);

  //gtk_wrapper_drawing_area_set_width(d->drawing_area, w, d->zoom);
  gtk_wrapper_drawing_area_set_dimension(d->drawing_area, d->allocated_width,
                                         d->allocated_height, d->zoom);
}

static void elem_calc_dimension(elem *e)
{
  int alteration_min_x = 0;
  int alteration_max_x = 0;
  switch(e->type) {
    case NOTE: {
      int dot_space;
      int min_x;
      int max_x;
      int min_y;
      int max_y;
      /* width/height/center of the note head */
      /* black and white heads have same dimensions, so no need to separate */
      min_x = -(images[I_BLACK_HEAD].width-1)/2;
      max_x = images[I_BLACK_HEAD].width/2 + NOTE_HEAD_SPACE;
      min_y = -images[I_BLACK_HEAD].height/2;
      max_y = (images[I_BLACK_HEAD].height-1)/2;
      /* calculate space taken by stem */
      switch (e->u.note.duration) {
        case RONDE:
          e->dim.stem_min_y = 0;
          e->dim.stem_max_y = 0;
          e->dim.stem_type = RONDE;
          break;
        case BLANCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = BLANCHE;
          break;
        case NOIRE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = NOIRE;
          break;
        case CROCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = CROCHE;
          break;
        case DCROCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2 + DCROCHE_ADJUST;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2 + DCROCHE_ADJUST;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2 - DCROCHE_ADJUST;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2 - DCROCHE_ADJUST;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = DCROCHE;
          break;
        case DDCROCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2 + DDCROCHE_ADJUST;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2 + DDCROCHE_ADJUST;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2 - DDCROCHE_ADJUST;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2 - DDCROCHE_ADJUST;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = DDCROCHE;
          break;
        case DDDCROCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2 + DDDCROCHE_ADJUST;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2 + DDDCROCHE_ADJUST;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2 - DDDCROCHE_ADJUST;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2 - DDDCROCHE_ADJUST;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = DDDCROCHE;
          break;
        case DDDDCROCHE:
          if (e->u.note.stem_down) {
            max_y = 7 * LINE_SPACING / 2 + DDDDCROCHE_ADJUST;
            e->dim.stem_max_y = 7 * LINE_SPACING / 2 + DDDDCROCHE_ADJUST;
            e->dim.stem_min_y = 0;
          } else {
            min_y = -7 * LINE_SPACING / 2 - DDDDCROCHE_ADJUST;
            e->dim.stem_min_y = -7 * LINE_SPACING / 2 - DDDDCROCHE_ADJUST;
            e->dim.stem_max_y = 0;
          }
          e->dim.stem_type = DDDDCROCHE;
          break;
      }
      /* add up space from alteration */
      switch (e->u.note.alteration) {
        case NORMAL: break;
        case DIESE: {
          int min_diese_y = -images[I_DIESE].height / 2;
          int max_diese_y = (images[I_DIESE].height-1) / 2;
//          min_x -= images[I_DIESE].width + DIESE_SPACE;
          alteration_min_x = min_x - (images[I_DIESE].width + DIESE_SPACE);
          alteration_max_x = min_x;
          if (min_diese_y < min_y) min_y = min_diese_y;
          if (max_diese_y > max_y) max_y = max_diese_y;
          break;
        }
        case BEMOL: {
          int min_bemol_y = -images[I_BEMOL].height / 2 - BEMOL_ADJUST;
          int max_bemol_y = (images[I_BEMOL].height-1) / 2 - BEMOL_ADJUST;
//          min_x -= images[I_BEMOL].width + BEMOL_SPACE;
          alteration_min_x = min_x - (images[I_BEMOL].width + BEMOL_SPACE);
          alteration_max_x = min_x;
          if (min_bemol_y < min_y) min_y = min_bemol_y;
          if (max_bemol_y > max_y) max_y = max_bemol_y;
          break;
        }
        case BECARRE: {
          int min_becarre_y = -images[I_BECARRE].height / 2;
          int max_becarre_y = (images[I_BECARRE].height-1) / 2;
//          min_x -= images[I_BECARRE].width + BECARRE_SPACE;
          alteration_min_x = min_x - (images[I_BECARRE].width + BECARRE_SPACE);
          alteration_max_x = min_x;
          if (min_becarre_y < min_y) min_y = min_becarre_y;
          if (max_becarre_y > max_y) max_y = max_becarre_y;
          break;
        }
      }
      /* dots' stuff (they get close to note's head, take care when
       * calculating new width when there is CROCHE stuff)
       * STRONG ASSUMPTION: a dot is supposed to have a small height
       * (ie. we don't look at its height to get the height of the symbol)
       */
      dot_space = e->u.note.nb_dots * (images[I_DOT].width + DOT_SPACE);
      if (e->u.note.nb_dots) dot_space += DOT_NOTE_SPACE;
//      max_x += dot_space;
      /* note head delta stuff */
      min_x += e->u.note.head_delta;
      max_x += e->u.note.head_delta;
      /* alteration delta stuff */
      if (e->u.note.alteration != NORMAL) {
        alteration_min_x += e->u.note.head_delta + e->u.note.alteration_delta;
        alteration_max_x += e->u.note.head_delta + e->u.note.alteration_delta;
        if (alteration_min_x < min_x) min_x = alteration_min_x;
        if (alteration_max_x > max_x) max_x = alteration_max_x;
      }
      /* dots delta stuff */
      if (e->u.note.nb_dots) {
        int dots_min_x = max_x + e->u.note.dots_delta;
        int dots_max_x = max_x + dot_space - 1 + e->u.note.dots_delta;
        if (dots_min_x < min_x) min_x = dots_min_x;
        if (dots_max_x > max_x) max_x = dots_max_x;
      }
      e->dim.max_x = max_x;
      e->dim.min_x = min_x;
      e->dim.max_y = max_y;
      e->dim.min_y = min_y;
      return;
    }
    case SILENCE: {
      int min_x;
      int max_x;
      int min_y;
      int max_y;
      switch (e->u.silence.duration) {
        case RONDE:
          min_x = -(images[I_SILENCE_RONDE].width-1)/2;
          max_x = images[I_SILENCE_RONDE].width/2 + SILENCE_RONDE_SPACE;
          min_y = 0;
          max_y = images[I_SILENCE_RONDE].height-1;
          break;
        case BLANCHE:
          min_x = -(images[I_SILENCE_BLANCHE].width-1)/2;
          max_x = images[I_SILENCE_BLANCHE].width/2 + SILENCE_BLANCHE_SPACE;
          min_y = -(images[I_SILENCE_BLANCHE].height-1);
          max_y = 0;
          break;
        case NOIRE:
          min_x = -(images[I_SILENCE_NOIRE].width-1)/2;
          max_x = images[I_SILENCE_NOIRE].width/2 + SILENCE_NOIRE_SPACE;
          min_y = -images[I_SILENCE_NOIRE].height/2;
          max_y = (images[I_SILENCE_NOIRE].height-1)/2;
          break;
        case CROCHE:
          min_x = -(images[I_SILENCE_CROCHE].width-1)/2;
          max_x = images[I_SILENCE_CROCHE].width/2 + SILENCE_CROCHE_SPACE;
          min_y = -images[I_SILENCE_CROCHE].height/2;
          max_y = (images[I_SILENCE_CROCHE].height-1)/2;
          break;
        case DCROCHE:
          min_x = -(images[I_SILENCE_DCROCHE].width-1)/2;
          max_x = images[I_SILENCE_DCROCHE].width/2 + SILENCE_DCROCHE_SPACE;
          min_y = -images[I_SILENCE_DCROCHE].height/2;
          max_y = (images[I_SILENCE_DCROCHE].height-1)/2;
          break;
        case DDCROCHE:
          min_x = -(images[I_SILENCE_DDCROCHE].width-1)/2;
          max_x = images[I_SILENCE_DDCROCHE].width/2 + SILENCE_DDCROCHE_SPACE;
          min_y = -images[I_SILENCE_DDCROCHE].height/2;
          max_y = (images[I_SILENCE_DDCROCHE].height-1)/2;
          break;
        case DDDCROCHE:
          min_x = -(images[I_SILENCE_DDDCROCHE].width-1)/2;
          max_x = images[I_SILENCE_DDDCROCHE].width/2
                  + SILENCE_DDDCROCHE_SPACE;
          min_y = -images[I_SILENCE_DDDCROCHE].height/2;
          max_y = (images[I_SILENCE_DDDCROCHE].height-1)/2;
          break;
        case DDDDCROCHE:
          min_x = -(images[I_SILENCE_DDDDCROCHE].width-1)/2;
          max_x = images[I_SILENCE_DDDDCROCHE].width/2
                  + SILENCE_DDDDCROCHE_SPACE;
          min_y = -images[I_SILENCE_DDDDCROCHE].height/2;
          max_y = (images[I_SILENCE_DDDDCROCHE].height-1)/2;
          break;
      }
      /* dot stuff
       * STRONG ASSUMPTION: a dot is supposed to have a small height
       * (ie. we don't look at its height to get the height of the symbol)
       */
      max_x += e->u.silence.nb_dots * (images[I_DOT].width + DOT_SPACE);
      if (e->u.silence.nb_dots) max_x += DOT_SILENCE_SPACE;
      e->dim.max_x = max_x;
      e->dim.min_x = min_x;
      e->dim.max_y = max_y;
      e->dim.min_y = min_y;
      /* no stem, put a dummy one */
      e->dim.stem_min_y = 0;
      e->dim.stem_max_y = 0;
      e->dim.stem_type = RONDE;
      return;
    }
  }
}

static void chord_calc_dimension(chord *c, int x)
{
  int i;
  int max_x;
  int min_x;
  int max_y;
  int min_y;
  int stem_up_min_y;
  int stem_up_max_y;
  int stem_up_max_x;
  note_duration stem_up_type;
  int stem_down_min_y;
  int stem_down_max_y;
  int stem_down_max_x;
  note_duration stem_down_type;
  int stem_up_valid = 0;
  int stem_down_valid = 0;
  elem *e;

  switch (c->type) {
    case MESURE_BREAK:
      c->dim.min_x = 0;
      c->dim.max_x = MESURE_MARKER_WIDTH + MESURE_MARKER_SPACE;
      c->dim.min_y = 0;
      c->dim.max_y = 4 * LINE_SPACING + 1;
      return;
    case TONALITE: {
      int space;
      int type;
      int adjust = 0;
      switch (c->u.tonalite.alteration) {
        case NORMAL: /* we can't come here */ return;
        case DIESE: type = I_DIESE; space = DIESE_SPACE; break;
        case BEMOL: type = I_BEMOL; space = BEMOL_SPACE;
                    adjust = BEMOL_ADJUST; break;
        case BECARRE: type = I_BECARRE; space = BECARRE_SPACE; break;
      }
      c->dim.min_x = -(images[type].width-1)/2;
      c->dim.max_x = images[type].width/2 + space;
      c->dim.min_y = -images[type].height / 2;
      c->dim.max_y = (images[type].height-1) / 2;
      c->dim.min_y +=  3 * LINE_SPACING - adjust
                       - c->u.tonalite.position * LINE_SPACING/2;
      c->dim.max_y +=  3 * LINE_SPACING - adjust
                       - c->u.tonalite.position * LINE_SPACING/2;
      return;
    }
    case CLEF: {
      int space;
      int type;
      int adjust = 0;
      switch (c->u.clef.type) {
        case CLEF_C: type = I_CLEF_C; space = CLEF_SPACE;
                     adjust = CLEF_C_ADJUST; break;
        case CLEF_F: type = I_CLEF_F; space = CLEF_SPACE;
                     adjust = CLEF_F_ADJUST; break;
        case CLEF_G: type = I_CLEF_G; space = CLEF_SPACE;
                     adjust = CLEF_G_ADJUST; break;
        case CLEF_G_8va: type = I_CLEF_G_8va; space = CLEF_SPACE;
                         adjust = CLEF_G_8va_ADJUST; break;
        case CLEF_G_8vb: type = I_CLEF_G_8vb; space = CLEF_SPACE;
                         adjust = CLEF_G_8vb_ADJUST; break;
      }
      c->dim.min_x = -(images[type].width-1)/2;
      c->dim.max_x = images[type].width/2 + space;
      c->dim.min_y = -images[type].height / 2;
      c->dim.max_y = (images[type].height-1) / 2;
      c->dim.min_y +=  3 * LINE_SPACING - adjust
                       - c->u.clef.position * LINE_SPACING/2;
      c->dim.max_y +=  3 * LINE_SPACING - adjust
                       - c->u.clef.position * LINE_SPACING/2;
      return;
    }
    case NOTES_CHORD: break;
  }

  if (c->u.chord.size) {
    e = c->u.chord.data;
    elem_calc_dimension(e);
    min_x = e->dim.min_x;
    max_x = e->dim.max_x;
    switch (e->type) {
      case NOTE:
        min_y = e->dim.min_y + 3 * LINE_SPACING
                - e->u.note.pitch * LINE_SPACING/2;
        max_y = e->dim.max_y + 3 * LINE_SPACING
                - e->u.note.pitch * LINE_SPACING/2;
        if (e->u.note.stem_down) {
          stem_down_min_y = e->dim.stem_min_y + 3 * LINE_SPACING
                            - e->u.note.pitch * LINE_SPACING/2;
          stem_down_max_y = e->dim.stem_max_y + 3 * LINE_SPACING
                            - e->u.note.pitch * LINE_SPACING/2;
          stem_down_type = e->dim.stem_type;
          stem_down_valid = 1;
          stem_up_min_y = 0;
          stem_up_max_y = 0;
          stem_up_type = RONDE;
        } else {
          stem_up_min_y = e->dim.stem_min_y + 3 * LINE_SPACING
                          - e->u.note.pitch * LINE_SPACING/2;
          stem_up_max_y = e->dim.stem_max_y + 3 * LINE_SPACING
                          - e->u.note.pitch * LINE_SPACING/2;
          stem_up_type = e->dim.stem_type;
          stem_up_valid = 1;
          stem_down_min_y = 0;
          stem_down_max_y = 0;
          stem_down_type = RONDE;
        }
        break;
      case SILENCE:
        min_y = e->dim.min_y + 3 * LINE_SPACING
                - e->u.silence.position * LINE_SPACING/2;
        max_y = e->dim.max_y + 3 * LINE_SPACING
                - e->u.silence.position * LINE_SPACING/2;
        stem_up_min_y = 0;
        stem_up_max_y = 0;
        stem_up_type = RONDE;
        stem_down_min_y = 0;
        stem_down_max_y = 0;
        stem_down_type = RONDE;
        break;
    }
  } else {
    /* dummy values */
    min_y = 0; max_y = 1;
    min_x = 0; max_x = 1;
    stem_up_min_y = 0; stem_up_max_y = 0; stem_up_type = RONDE;
    stem_down_min_y = 0; stem_down_max_y = 0; stem_down_type = RONDE;
  }

  for (i = 1; i < c->u.chord.size; i++) {
    int nmin_y;
    int nmax_y;
    int nstem_up_min_y;
    int nstem_up_max_y;
    int nstem_up_type;
    int nstem_up_valid = 0;
    int nstem_down_min_y;
    int nstem_down_max_y;
    int nstem_down_type;
    int nstem_down_valid = 0;
    e = c->u.chord.data+i;
    elem_calc_dimension(e);
    if (e->dim.min_x < min_x) min_x = e->dim.min_x;
    if (e->dim.max_x > max_x) max_x = e->dim.max_x;
    switch (e->type) {
      case NOTE:
        nmin_y = e->dim.min_y + 3 * LINE_SPACING
                 - e->u.note.pitch * LINE_SPACING/2;
        nmax_y = e->dim.max_y + 3 * LINE_SPACING
                 - e->u.note.pitch * LINE_SPACING/2;
        if (e->u.note.stem_down) {
          nstem_down_min_y = e->dim.stem_min_y + 3 * LINE_SPACING
                             - e->u.note.pitch * LINE_SPACING/2;
          nstem_down_max_y = e->dim.stem_max_y + 3 * LINE_SPACING
                             - e->u.note.pitch * LINE_SPACING/2;
          nstem_down_type = e->dim.stem_type;
          nstem_down_valid = 1;
        } else {
          nstem_up_min_y = e->dim.stem_min_y + 3 * LINE_SPACING
                           - e->u.note.pitch * LINE_SPACING/2;
          nstem_up_max_y = e->dim.stem_max_y + 3 * LINE_SPACING
                           - e->u.note.pitch * LINE_SPACING/2;
          nstem_up_type = e->dim.stem_type;
          nstem_up_valid = 1;
        }
        break;
      case SILENCE:
        nmin_y = e->dim.min_y + 3 * LINE_SPACING
                - e->u.silence.position * LINE_SPACING/2;
        nmax_y = e->dim.max_y + 3 * LINE_SPACING
                - e->u.silence.position * LINE_SPACING/2;
        nstem_up_min_y = 0;
        nstem_up_max_y = 0;
        nstem_up_type = RONDE;
        nstem_down_min_y = 0;
        nstem_down_max_y = 0;
        nstem_down_type = RONDE;
        break;
    }
    if (nmin_y < min_y) min_y = nmin_y;
    if (nmax_y > max_y) max_y = nmax_y;
    if (nstem_up_valid) {
      if (stem_up_valid) {
        if (nstem_up_min_y < stem_up_min_y) stem_up_min_y = nstem_up_min_y;
        if (nstem_up_max_y > stem_up_max_y) stem_up_max_y = nstem_up_max_y;
        if (nstem_up_type > stem_up_type) stem_up_type = nstem_up_type;
      } else {
        stem_up_min_y = nstem_up_min_y;
        stem_up_max_y = nstem_up_max_y;
        stem_up_type = nstem_up_type;
        stem_up_valid = 1;
      }
    }
    if (nstem_down_valid) {
      if (stem_down_valid) {
        if (nstem_down_min_y < stem_down_min_y)
          stem_down_min_y = nstem_down_min_y;
        if (nstem_down_max_y > stem_down_max_y)
          stem_down_max_y = nstem_down_max_y;
        if (nstem_down_type > stem_down_type)
          stem_down_type = nstem_down_type;
      } else {
        stem_down_min_y = nstem_down_min_y;
        stem_down_max_y = nstem_down_max_y;
        stem_down_type = nstem_down_type;
        stem_down_valid = 1;
      }
    }
  }
  /* if note heads have been moved, min_x and max_x may be wrong, clen up */
  /* we compare against stem min and max and take the union of everything */
  /* stem up min x = images[I_BLACK_HEAD].width/2-1 */
  if (stem_up_valid && min_x > images[I_BLACK_HEAD].width/2 - 1)
    min_x = images[I_BLACK_HEAD].width/2 - 1;
  /* stem down min x = -images[I_BLACK_HEAD].width/2 */
  if (stem_down_valid && min_x > -images[I_BLACK_HEAD].width/2)
    min_x = -images[I_BLACK_HEAD].width/2;
  /* calc stem max x which depends on the stem type */
  /* we test both stem directions */
  /* first up */
  switch (stem_up_type) {
    case RONDE: stem_up_max_x = 0; break;
    case BLANCHE: stem_up_max_x = 0; break;
    case NOIRE: stem_up_max_x = 0; break;
    case CROCHE: stem_up_max_x = images[I_UP_CROCHE].width - 1; break;
    case DCROCHE: stem_up_max_x = images[I_UP_DCROCHE].width - 1; break;
    case DDCROCHE: stem_up_max_x = images[I_UP_DDCROCHE].width - 1; break;
    case DDDCROCHE: stem_up_max_x = images[I_UP_DDDCROCHE].width - 1; break;
    case DDDDCROCHE: stem_up_max_x = images[I_UP_DDDDCROCHE].width - 1; break;
  }
  stem_up_max_x += images[I_BLACK_HEAD].width/2 + NOTE_HEAD_SPACE;
  if (max_x < stem_up_max_x) max_x = stem_up_max_x;
  /* then down */
  switch (stem_down_type) {
    case RONDE: stem_down_max_x = 0; break;
    case BLANCHE: stem_down_max_x = 0; break;
    case NOIRE: stem_down_max_x = 0; break;
    case CROCHE: stem_down_max_x = images[I_DOWN_CROCHE].width - 1; break;
    case DCROCHE: stem_down_max_x = images[I_DOWN_DCROCHE].width - 1; break;
    case DDCROCHE: stem_down_max_x = images[I_DOWN_DDCROCHE].width - 1; break;
    case DDDCROCHE:
      stem_down_max_x = images[I_DOWN_DDDCROCHE].width - 1; break;
    case DDDDCROCHE:
      stem_down_max_x = images[I_DOWN_DDDDCROCHE].width - 1; break;
  }
  stem_down_max_x += -images[I_BLACK_HEAD].width/2 + NOTE_HEAD_SPACE;
  if (max_x < stem_down_max_x) max_x = stem_down_max_x;
  c->dim.max_x = max_x;
  c->dim.min_x = min_x;
  c->dim.max_y = max_y;
  c->dim.min_y = min_y;
  c->dim.stem_up_min_y = stem_up_min_y;
  c->dim.stem_up_max_y = stem_up_max_y;
  c->dim.stem_up_type = stem_up_type;
  c->dim.stem_down_min_y = stem_down_min_y;
  c->dim.stem_down_max_y = stem_down_max_y;
  c->dim.stem_down_type = stem_down_type;
}

static note_duration chord_min_duration(chord *c)
{
  note_duration ret = RONDE;
  int i;

  for (i = 0; i < c->u.chord.size; i++)
    switch (c->u.chord.data[i].type) {
      case NOTE:
        if (c->u.chord.data[i].u.note.duration > ret)
          ret = c->u.chord.data[i].u.note.duration;
        break;
      case SILENCE:
        if (c->u.chord.data[i].u.silence.duration > ret)
          ret = c->u.chord.data[i].u.silence.duration;
        break;
    }

  return ret;
}

static void music_line_calc_dimension(line *l)
{
  int i;
  int chord_x = LEFT_LINE_BORDER + 4;

  l->dim.width = LEFT_LINE_BORDER;
  l->dim.min_y = 0;
  l->dim.max_y = 4 * LINE_SPACING + 1;

  for (i = 0; i < l->u.music.size; i++) {
    chord *c = l->u.music.data+i;
    chord_calc_dimension(c, chord_x);
    if (c->dim.min_y < l->dim.min_y) l->dim.min_y = c->dim.min_y;
    if (c->dim.max_y > l->dim.max_y) l->dim.max_y = c->dim.max_y;
    c->dim.width = c->dim.max_x-c->dim.min_x+1;
    /* TONALITE or CLEF followed or preceeded by notes eats more space */
    if (c->type == TONALITE || c->type == CLEF) {
      /* is there a note after? */
      if (i != l->u.music.size - 1 &&
          l->u.music.data[i+1].type == NOTES_CHORD)
        c->dim.width += TONALITE_OR_CLEF_NOTE_SPACE;
    }
    if (c->type == NOTES_CHORD) {
      int inc_width = 0;
      int inc_width_2;
      /* is there a TONALITE or CLEF after? */
      if (i != l->u.music.size - 1 &&
          (l->u.music.data[i+1].type == TONALITE ||
           l->u.music.data[i+1].type == CLEF))
        inc_width = TONALITE_OR_CLEF_NOTE_SPACE;
      /* a chord has a minimal duration, use it to increase its width */
      switch (chord_min_duration(c)) {
        case RONDE: inc_width_2 = RONDE_SPACE; break;
        case BLANCHE: inc_width_2 = BLANCHE_SPACE; break;
        case NOIRE: inc_width_2 = NOIRE_SPACE; break;
        case CROCHE: inc_width_2 = CROCHE_SPACE; break;
        case DCROCHE: inc_width_2 = DCROCHE_SPACE; break;
        case DDCROCHE: inc_width_2 = DDCROCHE_SPACE; break;
        case DDDCROCHE: inc_width_2 = DDDCROCHE_SPACE; break;
        case DDDDCROCHE: inc_width_2 = DDDDCROCHE_SPACE; break;
      }
      if (inc_width_2 > inc_width) inc_width = inc_width_2;
      c->dim.width += inc_width;
    }
    c->dim.x = chord_x;
    l->dim.width = chord_x + c->dim.width;
    chord_x += c->dim.width;
  }

  l->dim.width += RIGHT_LINE_BORDER;

  if (l->dim.width < MIN_LINE_WIDTH) l->dim.width = MIN_LINE_WIDTH;

  l->dim.height = l->dim.max_y - l->dim.min_y + 1 + LINE_LINE_SPACE;
  if (l->dim.height < LINE_HEIGHT + LINE_LINE_SPACE)
    l->dim.height = LINE_HEIGHT + LINE_LINE_SPACE;
}

/* This function is ugly because it uses the zoom.
 * But we have to do it this way because fonts may
 * not scale as we expect. Pango may use 'nearest' available
 * fonts for a given size, so we can't just calculate using
 * the font of zoom level 1 and then multiply by the zoom in use.
 * Yeah, welcome in reality.
 */
static void text_line_calc_dimension(line *l, doc *d)
{
  int width;
  int height;
  gtk_wrapper_string_size(d->fonts, d->zoom,
                          l->u.text.data,
                          &width, &height);
  l->dim.width = width + TEXT_LEFT_BORDER + TEXT_RIGHT_BORDER;
  l->dim.height =  height + TEXT_UPPER_MARGIN + TEXT_LOWER_MARGIN;
  l->dim.min_y = 0;
  l->dim.max_y = height-1;
}

static void line_calc_dimension(line *l, doc *d)
{
  switch (l->type) {
    case LINE_MUSIC: return music_line_calc_dimension(l);
    case LINE_TEXT: return text_line_calc_dimension(l, d);
  }
}

static void reset_all(doc *d)
{
  /* let's be barbarian */
  int i;

  d->width = 0;
  d->height = UPPER_MARGIN + LOWER_MARGIN;

  for (i = 0; i < d->size; i++)
    d->data[i].next_invalid = i-1;

  /* position first line */
  if (d->size) d->data[0].dim.y = UPPER_MARGIN;

  d->invalid_line = i-1;
  d->deleted_line = -1;
  d->last_positioned_line = 0;
}

static void line_changed(doc *d, int line_number)
{
  line *l = &d->data[line_number];
  dprintf("line_changed %d (for post_treat) (%d)\n",
          line_number, l->next_invalid);
  if (l->next_invalid == -2) {
    d->height -= l->dim.height;
    l->next_invalid = d->invalid_line;
    d->invalid_line = line_number;
    if (d->last_positioned_line > line_number)
      d->last_positioned_line = line_number;
  }
}

void size_delete_chord(doc *d, int line_number, int chord_number)
{
  line_changed(d, line_number);
}

void size_change_chord(doc *d, int line_number, int chord_number)
{
  line_changed(d, line_number);
}

void size_new_chord(doc *d, int line_number, int chord_number)
{
  line_changed(d, line_number);
}

void size_prenew_line(doc *d, int line_number)
{
  /* barbarianly force validation of lines, because one
   * line will be added so indexes will change
   * (bad design in the data structure or algorithms one
   * might argue and I agree, we'll see if this survives up
   * to the release or not)
   */
  size_post_treatment(d);
}

void size_new_line(doc *d, int line)
{
  /* we set up the 'next_invalid' for the new line */
  d->data[line].next_invalid = -2;
  /* its height is 0 */
  d->data[line].dim.height = 0;
  line_changed(d, line);

  if (line == 0) {
    /* if first line, it has to be positioned */
    d->last_positioned_line = 0;
    d->data[0].dim.y = UPPER_MARGIN;
  } else {
    /* otherwise the last positioned is the previous one */
    if (d->last_positioned_line > line - 1)
      d->last_positioned_line = line - 1;
  }
}

void size_presplit_line(doc *d, int line_number)
{
  /* barbarianly force validation of lines, because one
   * line will be added so indexes will change
   * (bad design in the data structure or algorithms one
   * might argue and I agree, we'll see if this survives up
   * to the release or not)
   */
  size_post_treatment(d);
}

void size_split_line(doc *d, int first_line)
{
  /* we set up the 'next_invalid' for the new line */
  d->data[first_line+1].next_invalid = -2;
  /* its height is 0 */
  d->data[first_line+1].dim.height = 0;
  line_changed(d, first_line);
  line_changed(d, first_line+1);
}

void size_premerge_line(doc *d, int line_number)
{
  /* barbarianly force validation of lines, because one
   * line will be deleted so indexes will change
   * (bad design in the data structure or algorithms one
   * might argue and I agree, we'll see if this survives up
   * to the release or not)
   *
   * we also will destroy the line, so its height has to be
   * correct, so if the code ever changes, don't forget to call
   * a line_calc_dimension on line_number
   */
  size_post_treatment(d);

  /* we have to change the height, the line will disappear */
  d->height -= d->data[line_number].dim.height;
}

void size_merge_line(doc *d, int line_number)
{
  line_changed(d, line_number);
}

void size_predelete_line(doc *d, int line_number)
{
  /* barbarianly force validation of lines, because one
   * line will be deleted so indexes will change
   * (bad design in the data structure or algorithms one
   * might argue and I agree, we'll see if this survives up
   * to the release or not)
   *
   * we also will destroy the line, so its height has to be
   * correct, so if the code ever changes, don't forget to call
   * a line_calc_dimension on line_number
   */
  size_post_treatment(d);

  /* we have to change the height, the line will disappear */
  d->height -= d->data[line_number].dim.height;
}

void size_delete_line(doc *d, int line_number)
{
  if (d->deleted_line == -1 || line_number < d->deleted_line)
    d->deleted_line = line_number;

  if (line_number == 0) {
    /* always at least one line so we are allowed to access data[0] */
    d->last_positioned_line = 0;
    d->data[0].dim.y = UPPER_MARGIN;
    return;
  }

  if (d->last_positioned_line > line_number - 1)
    d->last_positioned_line = line_number - 1;
}

void size_set_text(doc *d, int line)
{
  line_changed(d, line);
}

void size_reset(doc *d)
{
  reset_all(d);
}

void size_load(doc *d)
{
  reset_all(d);
}

void size_post_treatment(doc *d)
{
  int i;
  int first_line_treated = d->size;

  dprintf("size_post_treatment\n");
  /* calculate all invalidate lines */
  i = d->invalid_line;
  while (i >= 0) {
    int next;
    line_calc_dimension(d->data+i, d);
    dprintf("size_post_treatment treated line %d\n", i);
    if (i < first_line_treated) first_line_treated = i;
    d->height += d->data[i].dim.height;
    next = d->data[i].next_invalid;
    d->data[i].next_invalid = -2;
    i = next;
  }

  /* take into account deleted lines */
  if (d->deleted_line != -1) {
    if (d->deleted_line < first_line_treated)
      first_line_treated = d->deleted_line;
    if (first_line_treated == d->size)
      invalidate_next(d);
    else
      invalidate_line_and_next(d, first_line_treated);
  } else {
    /* case where there is no deleted line */
    if (first_line_treated != d->size)
      invalidate_line_and_next(d, first_line_treated);
  }

  d->invalid_line = -1;

  /* set vertical scrollbar (if needed) */
  if (d->height != d->current_height)  {
    d->current_height = d->height;
    //gtk_wrapper_drawing_area_set_height(d->drawing_area, d->height, d->zoom);

    setup_width_and_height(d);

    gtk_wrapper_drawing_area_set_dimension(d->drawing_area, d->allocated_width,
                                           d->allocated_height, d->zoom);
  }
}

void ensure_line_positioned(doc *d, int ln)
{
  line *l;
  int cur;

#if 0
  if (d->size == 0) return;

  if (d->last_line_valid == -1) {
    l = d->data;
    if (!l->size_valid) line_calc_dimension(l, d);
    l->dim.y = UPPER_MARGIN;
    d->last_line_valid = 0;
  }
#endif

  cur = d->last_positioned_line + 1;
  while (cur <= ln) {
    l = d->data+cur;
//    if (!l->size_valid) line_calc_dimension(l, d);
    l->dim.y = d->data[cur-1].dim.y + d->data[cur-1].dim.height;
    d->last_positioned_line++;
    cur++;
  }
}

void ensure_height_in_view(doc *d, int y_max)
{
  line *l;
  int cur;

  dprintf("ensure_height_valid y_max=%d last valid line=%d\n",
          y_max, d->last_positioned_line);

  cur = d->last_positioned_line + 1;
  while (cur < d->size) {
    l = d->data+cur;
    l->dim.y = d->data[cur-1].dim.y + d->data[cur-1].dim.height;
    d->last_positioned_line++;
    /* if the line is greater than what we need, we're done */
    if (l->dim.y > y_max) break;
    cur++;
  }
}

void get_crossed_lines(doc *d, int y_min, int y_max,
                       int *first_line, int *last_line)
{
  int min;
  int max;
  int cur;
  int tst;
  line *l;

  /* take care of upper margin */
  if (y_min < UPPER_MARGIN) {
    /* if both are < UPPER_MARGIN, we cross no line */
    if (y_max < UPPER_MARGIN) {
      *first_line = -1;
      return;
    }
    y_min = UPPER_MARGIN;
  }
  /* maybe we could avoid the +1 here, not thought of it */
  if (y_max < y_min + 1) y_max = y_min + 1;

  ensure_height_in_view(d, y_max);

  min = 0;
  max = d->last_positioned_line;

  /* look for the first line containing y_min */
  while (1) {
    if (max < min) {
      *first_line = -1;
      return;
    }
    cur = (max+min)/2;
    l = d->data + cur;
    /* if cur is inside then tst=0 else if cur < line then tst=1 else -1 */
    if (l->dim.y <= y_min && l->dim.y + l->dim.height > y_min) tst = 0;
    else if (l->dim.y > y_min) tst = 1;
    else tst = -1;
    if (tst > 0)
      max = cur-1;
    else if (tst < 0)
      min = cur+1;
    else {
      /* found, get last list line containing y_max */
      *first_line = cur;
      while (cur <= d->last_positioned_line && d->data[cur].dim.y <= y_max)
        cur++;
      *last_line = cur - 1;
      return;
    }
  }

  /* dummy return */
  return;
}

void get_crossed_chords(line *l, int x_min, int x_max,
                        int *first_chord, int *last_chord)
{
  int min;
  int max;
  int cur;
  int tst;
  chord *c;

  min = 0;
  max = l->u.music.size-1;

  /* take care of left margin */
  x_min -= LEFT_MARGIN;
  x_max -= LEFT_MARGIN;

  /* start from first chord (if any) */
  if (l->u.music.size && x_min < l->u.music.data[0].dim.x) {
    /* if both are < then no chord in view */
    if (x_max < l->u.music.data[0].dim.x) {
      *first_chord = -1;
      return;
    }
    x_min = l->u.music.data[0].dim.x;
  }
  if (x_max < x_min + 1) x_max = x_min + 1;

  /* look for the first chord containing x_min */
  while (1) {
    if (max < min) {
      *first_chord = -1;
      return;
    }
    cur = (max+min)/2;
    c = l->u.music.data + cur;
    /* if cur is inside then tst=0 else if cur < line then tst=1 else -1 */
    if (c->dim.x <= x_min && c->dim.x + c->dim.width > x_min) tst = 0;
    else if (c->dim.x > x_min) tst = 1;
    else tst = -1;
    if (tst > 0)
      max = cur-1;
    else if (tst < 0)
      min = cur+1;
    else {
      /* found, get last list line containing y_max */
      *first_chord = cur;
      while (cur <= l->u.music.size-1 && l->u.music.data[cur].dim.x <= x_max)
        cur++;
      *last_chord = cur - 1;
      return;
    }
  }

  /* dummy return */
  return;
}

void set_zoom(doc *d, int z)
{
  int i;
  if (d->zoom == z) return;
  dprintf("set zoom %d\n", z);
  d->zoom = z;
  invalidate_all(d);

  /* we have to force the recalculation of all text lines
   * (but it's a fucking bad hack to do it the way we do,
   * maybe we should use a list of text lines? maybe some
   * other mechanism?) (HASHTABLES! HASHTABLES!) (SHUT UP!
   * Keep quiet. Good boy.)
   */
  for (i = 0; i < d->size; i++)
    if (LINE_TYPE(d, i) == LINE_TEXT)
      line_changed(d, i);
  size_post_treatment(d);

  setup_width_and_height(d);

  gtk_wrapper_drawing_area_set_dimension(d->drawing_area,
                                         d->allocated_width,
                                         d->allocated_height,
                                         d->zoom);
  show_cursor(&d->cursor);
}
